unit UAssembler;

{$mode delphi}

interface

uses
  Classes, SysUtils, UInstructions, UCpu, Generics.Collections, StrUtils,
  UMemory, UMessages;

type
  TErrorEvent = procedure (Text: string; Pos: TPoint) of object;

  { TLabelRef }

  TLabelRef = record
    LabelName: string;
    RefPos: Integer;
    TextPos: TPoint;
    class function Create(LabelName: string; RefPos: Integer): TLabelRef; static;
  end;

  TTokenKind = (tkKeyword, tkString, tkNumber, tkSpecialSymbol, tkEof, tkIdentifier);

  TToken = record
    Kind: TTokenKind;
    Value: string;
    Pos: TPoint;
  end;

  { TParserPos }

  TParserPos = record
    Index: Integer;
    Pos: TPoint;
    procedure Reset;
    procedure NextChar;
    procedure NextLine;
  end;

  { TParser }

  TParser = class
  private
    FOnError: TErrorEvent;
  public
    Pos: TParserPos;
    Source: string;
    function IsDigit(Value: Char): Boolean;
    function IsSpecialSymbol(C: Char): Boolean;
    function IsAlpha(C: Char): Boolean;
    function IsAlphaNumeric(C: Char): Boolean;
    function IsWhiteSpace(C: Char): Boolean;
    function ReadNext: TToken;
    function CheckNext(Kind: TTokenKind; Value: string = ''): Boolean;
    function CheckNextKind(Kind: TTokenKind): Boolean;
    procedure Expect(Kind: TTokenKind; Value: string);
    procedure Error(Text: string; Pos: TPoint);
    constructor Create;
    property OnError: TErrorEvent read FOnError write FOnError;
  end;

  { TAssembler }

  TAssembler = class
  private
    FOnError: TErrorEvent;
    Parser: TParser;
    procedure UpdateLabelRefs;
    procedure ParseNumParam(Token: TToken);
  public
    InstructionSet: TInstructionSet;
    Memory: TMemory;
    Labels: TDictionary<string, Integer>;
    LabelRefs: TList<TLabelRef>;
    Messages: TMessages;
    procedure Error(Text: string; Pos: TPoint);
    procedure Compile(Source: string);
    procedure LoadFromFile(FileName: string);
    function ParseStr(var Text: string; Separator: string): string;
    constructor Create;
    destructor Destroy; override;
    property OnError: TErrorEvent read FOnError write FOnError;
  end;


implementation

{ TParserPos }

procedure TParserPos.Reset;
begin
  Index := 1;
  Pos := Point(1, 1);
end;

procedure TParserPos.NextChar;
begin
  Inc(Index);
  Inc(Pos.X);
end;

procedure TParserPos.NextLine;
begin
  Inc(Index);
  Pos.X := 1;
  Inc(Pos.Y);
end;

{ TParser }

function TParser.IsDigit(Value: Char): Boolean;
begin
  Result := Value in ['0'..'9'];
end;

function TParser.IsSpecialSymbol(C: Char): Boolean;
begin
  Result := (C = ':') or (C = ',');
end;

function TParser.IsAlpha(C: Char): Boolean;
begin
  Result := (C in ['a'..'z']) or (C in ['A'..'Z']);
end;

function TParser.IsAlphaNumeric(C: Char): Boolean;
begin
  Result := IsAlpha(C) or IsDigit(C) or (C = '_');
end;

function TParser.IsWhiteSpace(C: Char): Boolean;
begin
  Result := (C = ' ') or (C = #9);
end;

function TParser.ReadNext: TToken;
type
  TParserState = (psNone, psNumber, psString, psComment, psIdentifier);
var
  C: Char;
  State: TParserState;
begin
  State := psNone;
  Result.Value := '';
  while Pos.Index < Length(Source) do begin
    C := Source[Pos.Index];
    if State = psNone then begin
      if IsWhiteSpace(C) then begin
      end else
      if C = ';' then begin
        State := psComment;
      end else
      if IsAlpha(C) then begin
        Result.Pos := Pos.Pos;
        Result.Kind := tkIdentifier;
        Result.Value := C;
        State := psIdentifier;
      end else
      if IsDigit(C) then begin
        Result.Pos := Pos.Pos;
        Result.Kind := tkNumber;
        Result.Value := C;
        State := psNumber;
      end else
      if C = '''' then begin
        Result.Pos := Pos.Pos;
        Result.Kind := tkString;
        State := psString;
      end else
      if C = #10 then begin
        Pos.NextLine;
        Continue;
      end else
      if IsSpecialSymbol(C) then begin
        Result.Pos := Pos.Pos;
        Result.Kind := tkSpecialSymbol;
        Result.Value := C;
        Pos.NextChar;
        Break;
      end else
        Error('Unknown character ' + C, Pos.Pos);
    end else
    if State = psIdentifier then begin
      if IsAlphaNumeric(C) then begin
        Result.Value := Result.Value + C;
      end else begin
        Break;
      end;
    end else
    if State = psComment then begin
      if (C = #10) or (C = #13) then begin
        State := psNone;
        Continue;
      end;
    end else
    if State = psNumber then begin
      if IsDigit(C) then Result.Value := Result.Value + C
      else begin
        Break;
      end;
    end else
    if State = psString then begin
      if C = '''' then begin
        Pos.NextChar;
        Break;
      end else begin
        Result.Value := Result.Value + C;
      end;
    end;
    Pos.NextChar;
  end;
  if (State = psNone) and (Pos.Index >= Length(Source)) then begin
    Result.Kind := tkEof;
    Result.Value := '';
  end;
end;

function TParser.CheckNext(Kind: TTokenKind; Value: string = ''): Boolean;
var
  LastPos: TParserPos;
  Token: TToken;
begin
  LastPos := Pos;
  Token := ReadNext;
  Result := (Token.Kind = Kind) and (Token.Value = Value);
  Pos := LastPos;
end;

function TParser.CheckNextKind(Kind: TTokenKind): Boolean;
var
  LastPos: TParserPos;
  Token: TToken;
begin
  LastPos := Pos;
  Token := ReadNext;
  Result := Token.Kind = Kind;
  Pos := LastPos;
end;

procedure TParser.Expect(Kind: TTokenKind; Value: string);
var
  Token: TToken;
begin
  Token := ReadNext;
  if (Token.Kind <> Kind) or (Token.Value <> Value) then
    Error('Expected ' + Value + ' but ' + Token.Value +' found.', Token.Pos);
end;

procedure TParser.Error(Text: string; Pos: TPoint);
begin
  if Assigned(FOnError) then
    FOnError(Text, Pos);
end;

constructor TParser.Create;
begin
  Pos.Reset;
end;

{ TLabelRef }

class function TLabelRef.Create(LabelName: string; RefPos: Integer): TLabelRef;
begin
  Result.LabelName := LabelName;
  Result.RefPos := RefPos;
end;

{ TAssembler }

procedure TAssembler.UpdateLabelRefs;
var
  I: Integer;
  Addr: Integer;
begin
  for I := 0 to LabelRefs.Count - 1 do begin
    if Labels.TryGetValue(LabelRefs[I].LabelName, Addr) then
      Memory[LabelRefs[I].RefPos] := Addr
      else Error('Label ' + LabelRefs[I].LabelName + ' referenced but not defined.', LabelRefs[I].TextPos);
  end;
end;

procedure TAssembler.ParseNumParam(Token: TToken);
var
  Addr: TInteger;
begin
  if Token.Kind = tkNumber then begin
    if TryStrToInt(Token.Value, Addr) then begin
      Memory.Write(Addr);
    end;
  end else
  if Token.Kind = tkIdentifier then begin;
    if Labels.TryGetValue(Token.Value, Addr) then begin
      Memory.Write(Addr);
    end else begin
      LabelRefs.Add(TLabelRef.Create(Token.Value, Memory.Position));
      Memory.Write(0);
    end;
  end else Error('Unexpected token ' + Token.Value + '.', Token.Pos);
end;

procedure TAssembler.Error(Text: string; Pos: TPoint);
begin
  Messages.AddMessage(Text, Pos);
  if Assigned(FOnError) then
    FOnError(Text, Pos);
end;

procedure TAssembler.Compile(Source: string);
var
  I: Integer;
  J: Integer;
  Name: string;
  Value: string;
  Line: string;
  Addr: Integer;
  LabelName: string;
  LabelRef: TLabelRef;
  InstructionInfo: TInstructionInfo;
  Token: TToken;
begin
  Messages.Clear;
  Memory.Size := 0;
  Labels.Clear;
  LabelRefs.Clear;
  Parser.Source := Source;
  while not Parser.CheckNextKind(tkEof) do begin
    Token := Parser.ReadNext;
    if Parser.CheckNext(tkSpecialSymbol, ':') then begin
      Parser.Expect(tkSpecialSymbol, ':');
      LabelName := Token.Value;
      if not Labels.TryGetValue(LabelName, Addr) then begin
        Labels.Add(LabelName, Memory.Position);
      end else Error('Duplicate label ' + LabelName + '.', Token.Pos);
    end else
    if Token.Kind = tkIdentifier then begin
      if LowerCase(Token.Value) = 'db' then begin
        Token := Parser.ReadNext;
        if Token.Kind = tkString then begin
          for J := 1 to Length(Token.Value) do
            Memory.Write(Ord(Token.Value[J]));
        end else
          ParseNumParam(Token);
        Continue;
      end else begin
        InstructionInfo := InstructionSet.SearchName(Token.Value);
        if Assigned(InstructionInfo) then begin
          Memory.Write(Integer(InstructionInfo.Instruction));
          for J := 0 to Length(InstructionInfo.Params) - 1 do begin
            if J > 0 then
              Parser.Expect(tkSpecialSymbol, ',');
            Token := Parser.ReadNext;
            if InstructionInfo.Params[J] = ptReg then begin
              if (Token.Value <> '') and (Token.Value[1] = 'R') then
                Token.Value := Copy(Token.Value, 2, MaxInt)
                else Error('Expected register name.', Token.Pos);
              Memory.Write(StrToInt(Token.Value));
            end else
            if InstructionInfo.Params[J] = ptNumber then begin
              ParseNumParam(Token);
            end;
          end;
        end else Error('Unknown keyword ' + Token.Value, Token.Pos);
      end;
    end;
  end;
  UpdateLabelRefs;
  Error('Compilation finished.', Point(0, 0));
end;

procedure TAssembler.LoadFromFile(FileName: string);
var
  Lines: TStringList;
begin
  Lines := TStringList.Create;
  try
    Lines.LoadFromFile(FileName);
    Compile(Lines.Text);
  finally
    Lines.Free;
  end;
end;

function TAssembler.ParseStr(var Text: string; Separator: string): string;
var
  P: Integer;
begin
  P := Pos(Separator, Text);
  if P > 0 then begin
    Result := Trim(Copy(Text, 1, P - 1));
    Text := Trim(Copy(Text, P + 1, MaxInt));
  end else begin
    Result := Trim(Text);
    Text := '';
  end;
end;

constructor TAssembler.Create;
begin
  Parser := TParser.Create;
  Parser.OnError := Error;
  Messages := TMessages.Create;
  Memory := TMemory.Create;
  InstructionSet := TInstructionSet.Create;
  Labels := TDictionary<string, Integer>.Create;
  LabelRefs := TList<TLabelRef>.Create;
end;

destructor TAssembler.Destroy;
begin
  FreeAndNil(Labels);
  FreeAndNil(LabelRefs);
  FreeAndNil(InstructionSet);
  FreeAndNil(Memory);
  FreeAndNil(Messages);
  FreeAndNil(Parser);
  inherited;
end;

end.

