unit UCpu;

{$mode delphi}

interface

uses
  Classes, SysUtils;

type
  TInstruction = (inNop, inHalt, inSet, inCopy, inInc, inDec, inLoad, inStore,
    inAdd, inSub, inIn, inOut, inJump, inJumpZero, inJumpNotZero, inPush, inPop,
    inCall, inRet, inAnd, inOr, inXor, inShl, inShr, inMul, inDiv, inMod,
    inJumpRel);

  TInteger = Integer;

  TCpuThread = class;
  TOutputEvent = procedure (Port: TInteger; Value: TInteger) of object;
  TInputEvent = function (Port: TInteger): TInteger of object;

  { TCpu }

  TCpu = class
  private
    FRunning: Boolean;
    FThread: TCpuThread;
    FOnInput: TInputEvent;
    FOnOutput: TOutputEvent;
    FNextInterupt: TInteger;
    FHalted: Boolean;
    procedure SetRunning(AValue: Boolean);
    procedure CheckInterreupts;
  public
    Ticks: Integer;
    IP: TInteger;
    SP: TInteger;
    Data: array of TInteger;
    R: array of TInteger;
    function ReadNext: TInteger;
    procedure WriteNext(Value: TInteger);
    procedure Step;
    procedure Start;
    procedure Stop;
    procedure Reset;
    procedure Interrupt(Index: TInteger);
    constructor Create;
    destructor Destroy; override;
    property OnInput: TInputEvent read FOnInput write FOnInput;
    property OnOutput: TOutputEvent read FOnOutput write FOnOutput;
    property Running: Boolean read FRunning write SetRunning;
    property Halted: Boolean read FHalted;
  end;

  { TCpuThread }

  TCpuThread = class(TThread)
    Cpu: TCpu;
    procedure Execute; override;
  end;


implementation

{ TCpuThread }

procedure TCpuThread.Execute;
begin
  repeat
    if not Cpu.Halted then Cpu.Step
      else Sleep(1);
    Cpu.CheckInterreupts;
  until Terminated;
end;

{ TCpu }

procedure TCpu.SetRunning(AValue: Boolean);
begin
  if FRunning = AValue then Exit;
  if AValue then Start
    else Stop;
end;

procedure TCpu.CheckInterreupts;
begin
  if FNextInterupt <> -1 then begin
    Dec(SP);
    Data[SP] := IP;
    IP := Data[FNextInterupt];
    FNextInterupt := -1;
    FHalted := False;
  end;
end;

function TCpu.ReadNext: TInteger;
begin
  if IP >= Length(Data) then IP := 0;
  Result := Data[IP];
  Inc(IP);
end;

procedure TCpu.WriteNext(Value: TInteger);
begin
  if IP >= Length(Data) then IP := 0;
  Data[IP] := Value;
  Inc(IP);
end;

procedure TCpu.Step;
var
  Instruction: TInstruction;
  Address: TInteger;
  Index: TInteger;
  Port: TInteger;
begin
  Instruction := TInstruction(ReadNext);
  case Instruction of
    inNop: ;
    inHalt: FHalted := True;
    inSet: R[ReadNext] := ReadNext;
    inCopy: R[ReadNext] := R[ReadNext];
    inLoad: R[ReadNext] := Data[R[ReadNext]];
    inStore: Data[R[ReadNext]] := R[ReadNext];
    inInc: Inc(R[ReadNext]);
    inDec: Dec(R[ReadNext]);
    inAdd: begin
      Index := ReadNext;
      R[Index] := R[Index] + R[ReadNext];
    end;
    inSub: begin
      Index := ReadNext;
      R[Index] := R[Index] - R[ReadNext];
    end;
    inIn: begin
      Index := ReadNext;
      Port := R[ReadNext];
      if Assigned(FOnInput) then R[Index] := FOnInput(Port);
    end;
    inOut: begin
      Port := R[ReadNext];
      if Assigned(FOnOutput) then FOnOutput(Port, R[ReadNext]);
    end;
    inJump: IP := ReadNext;
    inJumpRel: begin
      Address := ReadNext;
      IP := IP + Address;
    end;
    inJumpZero: begin
      Index := ReadNext;
      Address := ReadNext;
      if R[Index] = 0 then IP := Address;
    end;
    inJumpNotZero: begin
      Index := ReadNext;
      Address := ReadNext;
      if R[Index] <> 0 then IP := Address;
    end;
    inPush: begin
      Dec(SP);
      Data[SP] := R[ReadNext];
    end;
    inPop: begin
      R[ReadNext] := Data[SP];
      Inc(SP);
    end;
    inCall: begin
      Address := ReadNext;
      Dec(SP);
      Data[SP] := IP;
      IP := Address;
    end;
    inRet: begin
      IP := Data[SP];
      Inc(SP);
    end;
    inAnd: begin
      Index := ReadNext;
      R[Index] := R[Index] and R[ReadNext];
    end;
    inOr: begin
      Index := ReadNext;
      R[Index] := R[Index] or R[ReadNext];
    end;
    inXor: begin
      Index := ReadNext;
      R[Index] := R[Index] xor R[ReadNext];
    end;
    inShl: begin
      Index := ReadNext;
      R[Index] := R[Index] shl R[ReadNext];
    end;
    inShr: begin
      Index := ReadNext;
      R[Index] := R[Index] shr R[ReadNext];
    end;
    inMul: begin
      Index := ReadNext;
      R[Index] := R[Index] * R[ReadNext];
    end;
    inDiv: begin
      Index := ReadNext;
      R[Index] := R[Index] div R[ReadNext];
    end;
    inMod: begin
      Index := ReadNext;
      R[Index] := R[Index] mod R[ReadNext];
    end;
  end;
  Inc(Ticks);
end;

procedure TCpu.Start;
begin
  if not Running then begin
    Reset;
    FThread := TCpuThread.Create(True);
    FThread.Cpu := Self;
    FThread.Start;
    FRunning := True;
  end;
end;

procedure TCpu.Stop;
begin
  if Running then begin
    FHalted := True;
    FThread.Terminate;
    FThread.WaitFor;
    FreeAndNil(FThread);
    FRunning := False;
  end;
end;

procedure TCpu.Reset;
var
  I: Integer;
begin
  FNextInterupt := -1;
  FHalted := False;
  Ticks := 0;
  IP := Data[0]; // Reset interrupt vector
  SP := Length(Data);
  for I := 0 to Length(R) - 1 do
    R[I] := 0;
end;

procedure TCpu.Interrupt(Index: TInteger);
begin
  FNextInterupt := Index;
end;

constructor TCpu.Create;
begin
  SetLength(R, 16);
  SetLength(Data, 0);
end;

destructor TCpu.Destroy;
begin
  Stop;
  inherited;
end;

end.

