unit UMachine;

{$mode delphi}

interface

uses
  Classes, SysUtils, UCpu, Syncobjs, UMemory, Generics.Collections;

type
  TInputPort = (ipConsoleReadChar, ipConsoleInputCount, ipKeyboardRead,
    ipScreenGetWidth, ipScreenGetHeight, ipStorageGetSize, ipStorageRead);
  TOutputPort = (opConsoleWriteChar, opScreenSetAddr, opScreenWrite,
    opStorageSetAddr, opStorageWrite);

  TDevice = class
    Cpu: TCpu;
    InterruptVector: Integer;
  end;

  { TConsole }

  TConsole = class(TDevice)
    Lock: TCriticalSection;
    InputBuffer: TQueue<TInteger>;
    OutputBuffer: TQueue<TInteger>;
    constructor Create;
    destructor Destroy; override;
  end;

  { TScreen }

  TScreen = class(TDevice)
  private
    FSize: TPoint;
    procedure SetSize(AValue: TPoint);
  public
    Address: Integer;
    Data: array of Byte;
    Modified: Boolean;
    constructor Create;
    property Size: TPoint read FSize write SetSize;
  end;

  { TKeyboard }

  TKeyboard = class(TDevice)
    function ReadKey: TInteger;
  end;

  TStorage = class(TDevice)
  public
    F: TFileStream;
    FileName: string;
  end;

  { TMachine }

  TMachine = class
  private
    procedure OutputHandler(Port, Data: TInteger);
    function InputHandler(Port: TInteger): TInteger;
  public
    Memory: TMemory;
    Cpu: TCpu;
    Screen: TScreen;
    Keyboard: TKeyboard;
    Console: TConsole;
    Storage: TStorage;
    procedure PowerOn;
    procedure PowerOff;
    constructor Create;
    destructor Destroy; override;
  end;


implementation

{ TConsole }

constructor TConsole.Create;
begin
  Lock := TCriticalSection.Create;
  InputBuffer := TQueue<TInteger>.Create;
  OutputBuffer := TQueue<TInteger>.Create;
end;

destructor TConsole.Destroy;
begin
  InputBuffer.Free;
  OutputBuffer.Free;
  Lock.Free;
  inherited;
end;

{ TScreen }

procedure TScreen.SetSize(AValue: TPoint);
begin
  if FSize = AValue then Exit;
  FSize := AValue;
  SetLength(Data, FSize.X * FSize.Y);
end;

constructor TScreen.Create;
begin
  Size := Point(640, 480);
end;

{ TKeyboard }

function TKeyboard.ReadKey: TInteger;
begin
  Result := 0;
end;

{ TMachine }

procedure TMachine.OutputHandler(Port, Data: TInteger);
begin
  case TOutputPort(Port) of
    opConsoleWriteChar: begin
      Console.Lock.Acquire;
      try
        Console.OutputBuffer.Enqueue(Data);
      finally
        Console.Lock.Release;
      end;
    end;
    opScreenSetAddr: Screen.Address := Data;
    opScreenWrite: if (Screen.Address >= 0) and (Screen.Address < Length(Screen.Data)) then begin
      Screen.Data[Screen.Address] := Data;
      Inc(Screen.Address);
      Screen.Modified := True;
    end;
    opStorageSetAddr: Storage.F.Position := Data * SizeOf(TInteger);
    opStorageWrite: begin
      Storage.F.WriteBuffer(Data, SizeOf(TInteger));
    end;
  end;
end;

function TMachine.InputHandler(Port: TInteger): TInteger;
begin
  case TInputPort(Port) of
    ipConsoleReadChar: begin
      Console.Lock.Acquire;
      try
        if Console.InputBuffer.Count > 0 then begin
          Result := Console.InputBuffer.Dequeue;
        end else Result := 0;
      finally
        Console.Lock.Release;
      end;
    end;
    ipConsoleInputCount: begin
      Console.Lock.Acquire;
      try
        Result := Console.InputBuffer.Count;
      finally
        Console.Lock.Release;
      end;
    end;
    ipKeyboardRead: Result := Keyboard.ReadKey;
    ipScreenGetWidth: Result := Screen.Size.X;
    ipScreenGetHeight: Result := Screen.Size.Y;
    ipStorageGetSize: Result := Storage.F.Size div 4;
    ipStorageRead: begin
      Storage.F.ReadBuffer(Result, SizeOf(TInteger));
    end;
  end;
end;

procedure TMachine.PowerOn;
begin
  Cpu.Data := Memory.Data;
  Cpu.Start;
end;

procedure TMachine.PowerOff;
begin
  Cpu.Stop;
end;

constructor TMachine.Create;
begin
  Memory := TMemory.Create;
  Memory.Size := 10000;
  Cpu := TCpu.Create;
  Cpu.OnInput := InputHandler;
  Cpu.OnOutput := OutputHandler;
  Keyboard := TKeyboard.Create;
  Keyboard.Cpu := Cpu;
  Screen := TScreen.Create;
  Screen.Cpu := Cpu;
  Console := TConsole.Create;
  Console.Cpu := Cpu;
  Console.InterruptVector := 1;
  Storage := TStorage.Create;
  Storage.Cpu := Cpu;
end;

destructor TMachine.Destroy;
begin
  PowerOff;
  FreeAndNil(Storage);
  FreeAndNil(Console);
  FreeAndNil(Screen);
  FreeAndNil(Keyboard);
  FreeAndNil(Cpu);
  FreeAndNil(Memory);
  inherited;
end;

end.

